package radioai

import (
	"encoding/json"
	"log"
	"path/filepath"
	"strings"
	"sync"
	"time"

	"git.autistici.org/ale/radioai/masterelection"
	"github.com/coreos/go-etcd/etcd"
)

var (
	masterElectionPath = "/icecast/cluster/master"
	mountPrefix        = "/icecast/mounts/"
	nodePrefix         = "/icecast/nodes/"
)

// A mountpoint for a stream.
type Mount struct {
	// Name (path to the mountpoint).
	Name string

	// Username for source authentication.
	Username string

	// Password for source authentication.
	Password string

	// Fallback stream name (optional).
	Fallback string
}

// In-memory representation of the overall configuration (basically
// just a list of the known mounts).
type ClusterConfig struct {
	mounts map[string]*Mount
	lock   sync.Mutex
}

func NewClusterConfig() *ClusterConfig {
	return &ClusterConfig{
		mounts: make(map[string]*Mount),
	}
}

func (c *ClusterConfig) GetMount(name string) *Mount {
	c.lock.Lock()
	defer c.lock.Unlock()
	return c.mounts[name]
}

func (c *ClusterConfig) UpdateMount(m *Mount) {
	c.lock.Lock()
	defer c.lock.Unlock()
	c.mounts[m.Name] = m
}

func (c *ClusterConfig) DelMount(name string) {
	c.lock.Lock()
	defer c.lock.Unlock()
	delete(c.mounts, name)
}

func (c *ClusterConfig) ListMounts() []*Mount {
	c.lock.Lock()
	defer c.lock.Unlock()
	result := make([]*Mount, 0, len(c.mounts))
	for _, m := range c.mounts {
		result = append(result, m)
	}
	return result
}

type ConfigWatcher struct {
	client *etcd.Client
	config *ClusterConfig
	rch    chan *etcd.Response
	upch   chan bool
}

func NewConfigWatcher(client *etcd.Client, config *ClusterConfig, upch chan bool) *ConfigWatcher {
	return &ConfigWatcher{
		client: client,
		config: config,
		rch:    make(chan *etcd.Response, 100),
		upch:   upch,
	}
}

func (w *ConfigWatcher) syncer() {
	for response := range w.rch {
		mountName := filepath.Base(response.Key)
		if response.Action == "DELETE" {
			log.Printf("deleted mount '%s'", mountName)
			w.config.DelMount(mountName)
		} else {
			log.Printf("update to mount '%s'", mountName)
			var m Mount
			if err := json.NewDecoder(strings.NewReader(response.Value)).Decode(&m); err != nil {
				log.Printf("corrupted data: %s", err)
			} else {
				w.config.UpdateMount(&m)
			}
		}

		// Only flip the update signal once.
		select {
		case w.upch <- true:
		}
	}
}

func (w *ConfigWatcher) Run() {
	go w.syncer()

	// Run until the first successful Get().
	var index uint64
	for {
		responses, err := w.client.Get(mountPrefix)
		if err == nil {
			// Inject all the replies into the channel.
			for _, r := range responses {
				index = r.Index
				w.rch <- r
			}
			break
		}
		log.Printf("Get error: %s", err)
		time.Sleep(1 * time.Second)
	}

	// Now start the watcher.
	_, err := w.client.Watch(mountPrefix, index, w.rch, nil)
	if err != nil {
		log.Printf("Watch error: %s", err)
	}
}

type RadioCluster struct {
	ip          string
	client      *etcd.Client
	me          *masterelection.MasterElection
	config      *ClusterConfig
	watcher     *ConfigWatcher
	icecast     *IcecastController
	upch        chan bool
	livenessTtl uint64
}

func NewRadioCluster(ip string, client *etcd.Client) *RadioCluster {
	config := NewClusterConfig()

	// Network updates trigger icecast reconfiguration.
	upch := make(chan bool, 1)

	// MasterElection changes trigger an update.
	mech := make(chan int)
	go func() {
		for _ = range mech {
			select {
			case upch <- true:
			}
		}
	}()

	return &RadioCluster{
		ip:     ip,
		client: client,
		me: masterelection.NewMasterElection(
			client,
			masterElectionPath,
			ip,
			5,
			mech),
		config:      config,
		watcher:     NewConfigWatcher(client, config, upch),
		icecast:     NewIcecastController(ip),
		upch:        upch,
		livenessTtl: 2,
	}
}

func (rc *RadioCluster) presence() {
	for {
		if _, err := rc.client.Set(nodePrefix + rc.ip, rc.ip, rc.livenessTtl); err != nil {
			log.Printf("Set() error: %s", err)
			time.Sleep(100 * time.Millisecond)
		} else {
			time.Sleep(time.Duration(rc.livenessTtl / 2) * time.Second)
		}
	}
}

// Return the list of currently active nodes.
func (rc *RadioCluster) GetNodes() ([]string, error) {
	response, err := rc.client.Get(nodePrefix)
	if err != nil {
		return nil, err
	}
	result := make([]string, 0, len(response))
	for _, r := range response {
		result = append(result, r.Value)
	}
	return result, nil
}

func (rc *RadioCluster) Run() {
	// Bootstrap the config watcher. This ensures that we have a
	// full configuration (thanks to the Get() call) before we
	// start managing the icecast server.
	rc.watcher.Run()

	// Start the presence heartbeat.
	go rc.presence()

	for _ = range rc.upch {
		if err := rc.icecast.Update(rc.config, rc.me.IsMaster(), rc.me.MasterAddr); err != nil {
			log.Printf("Update() failed: %s", err)
		}
	}
}
